/*

Code extending the causal state reconstruction algorithm
technique exposed in "Automatic Filters for the Detection of Coherent
Structure in Spatiotemporal Systems". The algorithm below can
handle incremental estimation of the states as light cones
become available. Convergence is improved by running through the
state estimates in decreasing complexity. Also, states do really
handle all observations from their clustered past light cones.

The CausalStateAnalyzer class is generic and may easily be adapted.
Feed it light cones incrementally as you produce them, and it will
update the causal states and return you the best state estimate so far.

This code links to statistical testing routines that were developped
for the CSSR project (see Test.h and Test.cpp for proper credits).

Nicolas Brodu,
Released under GNU Public Licence v2 or above.
*/

#include <map>
#ifndef NO_TR1
#include <tr1/unordered_map>
#endif
#include <vector>

#include <assert.h>
#include <math.h>

#include "Test.h"

template<class PastLightCone, class FutureLightCone>
struct CausalStateAnalyzer {

    double significanceLevel;

    struct Cluster;

    struct FutureDistribution : public std::map<FutureLightCone, int> {
        int total;
        // the cluster this distribution belongs to.
        Cluster* cluster;

        FutureDistribution() : total(0), cluster(0) {}
        FutureDistribution(FutureLightCone flc) : total(0), cluster(0) {
            insert(flc);
        }

        int insert(FutureLightCone flc) {
            ++(*this)[flc];
            return ++total;
        }

        int remove(FutureLightCone flc) {
            typename std::map<FutureLightCone, int>::iterator flcIt = this->find(flc);
            assert(flcIt!=this->end()); // can't remove a future that was not inserted
            if (--flcIt->second == 0) this->erase(flcIt); // do not keep empty entries
            return --total; // may now be null
        }

        typedef typename std::map<FutureLightCone, int>::const_iterator const_iterator;

        void add(const FutureDistribution& distribution) {
            typename FutureDistribution::const_iterator it2 = distribution.begin();
            typename FutureDistribution::const_iterator eit2 = distribution.end();
            typename FutureDistribution::iterator it1 = this->lower_bound(it2->first);
            typename FutureDistribution::iterator eit1 = this->end();
            while (it2!=eit2) {
                // catchup on distribution
                while (it1!=eit1 && it1->first < it2->first) ++it1;
                // use amortized constant-time insert-with-hint function
                while (it2!=eit2 && (it1==eit1 || it2->first < it1->first)) {
                    it1 = ((std::map<FutureLightCone, int>*)this)->insert(it1, *it2++);
                    // eit1 = this->end(); // map property: insert does not invalidate iterators
                }
                // now elements are equal
                while (it2!=eit2 && it1!=eit1 && it2->first == it1->first) {
                    it1->second += it2->second;
                    ++it1; ++it2;
                }
            }
            total += distribution.total;
        }

        void sub(const FutureDistribution& distribution) {
            typename FutureDistribution::const_iterator it2 = distribution.begin();
            typename FutureDistribution::const_iterator eit2 = distribution.end();
            typename FutureDistribution::iterator it1 = this->lower_bound(it2->first);
            typename FutureDistribution::iterator eit1 = this->end();
            while (it2!=eit2) {
                // catchup on distribution
                #ifdef NDEBUG // assume correctness
                while (it1->first < it2->first) ++it1;
                #else // assertion mode
                while (it1!=eit1 && it1->first < it2->first) ++it1;
                // user can't remove a dist that was not added: there is NO it2 not in it1, elements are equal
                assert(it1!=eit1 && it1->first == it2->first);
                #endif
                // now elements are equal
                // so, subtract, and potentially erase if count is null
                int newCount = it1->second - it2->second;
                assert(newCount>=0); // user can't remove a distribution that wasn't added
                if (newCount==0) {
                    typename FutureDistribution::iterator it = it1; ++it;
                    this->erase(it1);
                    it1 = it;
                } else {
                    it1->second = newCount;
                    ++it1;
                }
                ++it2;
            }
            total -= distribution.total;
            assert(total>=0); // user can't remove a distribution that wasn't added
            // total maybe 0, but then, the cluster will be destroyed
        }

    };

    // Converts the distributions into a numerical format more appropriate for statistical testing
    struct DistributionTester : public Test {

        DistributionTester(bool isChiElseKS = true) : Test(isChiElseKS) {}

        // test if the 2 distribution match. Return value for significance testing
        double match(const FutureDistribution& dist1, const FutureDistribution& dist2) {
            std::vector<double> data1, data2;
            // create a bin for each distinct entry in either of the distributions
            // use the fact that std::map is sorted
            typename FutureDistribution::const_iterator it1 = dist1.begin();
            typename FutureDistribution::const_iterator it2 = dist2.begin();
            typename FutureDistribution::const_iterator eit1 = dist1.end();
            typename FutureDistribution::const_iterator eit2 = dist2.end();
            while (it1!=eit1 || it2!=eit2) {
                if ((it1!=eit1) && (it2==eit2 || it1->first<it2->first)) {
                    data1.push_back(it1->second); ++it1;
                    data2.push_back(0);
                }
                else if ((it2!=eit2) && (it1==eit1 || it2->first<it1->first)) {
                    data1.push_back(0);
                    data2.push_back(it2->second); ++it2;
                }
                else {
                    data1.push_back(it1->second); ++it1;
                    data2.push_back(it2->second); ++it2;
                }
            }
            // call parent test on the distributions. Both vectors have the same size by construction
            return RunTest(data1, dist1.total, data2, dist2.total, data1.size());
        }
    };

    DistributionTester tester;

    // A cluster contains a an average future distribution,
    // mapping between pasts to cluster is external
    struct Cluster {

        // The average distribution
        FutureDistribution distribution;
        // The number of times this cluster was observed
        int ninstances;

        // cluster complexity is -log(ninstances/totalInstances);
        // keep log(ninstances) since this is only modified when changing dist
        // then log(totalInstances) is computed only once for all clusters, and
        // a simple subtraction gives the complexity (instead of a log)
        double logNinstances;

        Cluster(FutureDistribution* dist) : ninstances(0), prev(0), next(0) {
            distribution.cluster = this;
            merge(dist);
        }

        void merge(FutureDistribution* dist) {
            assert(dist->cluster == 0);
            distribution.add(*dist);
            ninstances += dist->total;
            logNinstances = log2(ninstances);
            dist->cluster = this;
        }

        void remove(FutureDistribution* dist) {
            distribution.sub(*dist);
            ninstances -= dist->total;
            assert(ninstances>=0); // and log = -inf doesn't matter for destroyed cluster
            logNinstances = log2(ninstances);
            dist->cluster = 0;
        }

        // maintains a double-linked list of sorted clusters
        // O(Ncluster) but OK since add/remove an observation is O(Ncluster) anyway
        Cluster* prev;
        Cluster* next;
    };

    Cluster* firstCluster;
    Cluster* lastCluster;

#ifndef NO_TR1
    typedef std::tr1::unordered_map<PastLightCone,FutureDistribution*> DistributionMap;
#else
    typedef std::map<PastLightCone,FutureDistribution*> DistributionMap;
#endif
    DistributionMap observedDistributions;

    int totalInstances;
    // subtract each cluster logNinstances from that to get that cluster complexity
    double logTotalInstances;


    typedef int (FutureDistribution::*insert_or_remove_function)(FutureLightCone);

    template<insert_or_remove_function insert_or_remove, int deltaObs> Cluster* add_or_remove_Observation(PastLightCone plc, FutureLightCone flc) {

        // Step 1: find distribution for that past, if it exists
        typename DistributionMap::iterator distIt = observedDistributions.find(plc);

        FutureDistribution* observedDistribution;

        Cluster* oldCluster = 0;

        // found: remove the distribution from its cluster and update both
        if (distIt != observedDistributions.end()) {

            observedDistribution = distIt->second;
            Cluster* cluster = observedDistribution->cluster;

            // update cluster distrib. This may change the ordering
            cluster->remove(observedDistribution);

            // cluster is either empty or not at its place => should be removed from list
            bool wasMoved = false;
            if (cluster->ninstances==0
            || (cluster->prev!=0 && cluster->ninstances < cluster->prev->ninstances)
            || (cluster->next!=0 && cluster->ninstances > cluster->next->ninstances)) {
                // either maintain sorted linkage, or maintain first/last elements
                if (cluster->prev) cluster->prev->next = cluster->next;
                else firstCluster = cluster->next;
                if (cluster->next) cluster->next->prev = cluster->prev;
                else lastCluster = cluster->prev;
                wasMoved = true;
            }

            // destroy the cluster if it has become empty
            if (cluster->ninstances==0) {
                delete cluster; cluster = 0;
            }
            // or re-insert it at its new place if necessary
            else if (wasMoved) {
                // re-insert it at its place
                if (!firstCluster) {
                    cluster->prev = cluster->next = 0;
                    lastCluster = firstCluster = cluster;
                }
                else if (cluster->ninstances < firstCluster->ninstances) {
                    cluster->next = firstCluster;
                    cluster->prev = 0;
                    firstCluster->prev = cluster;
                    firstCluster = cluster;
                }
                else {
                    Cluster* c = firstCluster;
                    for (; c->next!=0; c = c->next) if (cluster->ninstances < c->next->ninstances) break;
                    cluster->next = c->next;
                    cluster->prev = c;
                    if (c->next) c->next->prev = cluster;
                    else lastCluster = cluster;
                    c->next = cluster;
                }
            }

            // now update the observedDistribution
            if ((observedDistribution->*insert_or_remove)(flc) == 0) {
                // that past has no future anymore... cleanup and done!
                observedDistributions.erase(distIt);
                delete observedDistribution;
                return cluster;
            }

            oldCluster = cluster;
        }
        // create a new distribution if this plc was not found, ensure it was found if we're removing an observation
        else {
            assert(deltaObs>0);
            observedDistribution = new FutureDistribution(flc);
            observedDistributions[plc] = observedDistribution;
        }

        // At this point, we get an updated observedDistribution (whether by adding or removing an flc)
        // Find the new cluster for that non-null observedDistribution
        Cluster* cluster = 0;

        // Step 2: run through the clusters from max complexity to low, take first match.
        // Idea: if 2 clusters pass the stat. test, take the more complex one.
        // this forces high-complexity clusters to be really different hence enhance the contrast.
        // conversely the low-complexity are better merged as their elements get reevaluated after the initial phase.
        bool shouldTestOldCluster = false;
        for (Cluster* c = firstCluster; c!=0; c = c->next) {
            if (c == oldCluster) {shouldTestOldCluster = true; continue;}
            if (tester.match(*observedDistribution, c->distribution) < significanceLevel) {
                cluster = c;
                break;
            }
        }
        if ((cluster==0) && shouldTestOldCluster) {
            if (tester.match(*observedDistribution, oldCluster->distribution) < significanceLevel) {
                cluster = oldCluster;
            }
        }

        // Step 3: Insert the distribution in the cluster just found, or create a new
        // one. This may be or not the old cluster. Update the cluster position in the
        // sorted cluster set.

        // no cluster found, create a new one
        bool wasMoved = false;
        if (cluster == 0) {
            cluster = new Cluster(observedDistribution);
            wasMoved = true;
        }
        // or update it
        else {
            cluster->merge(observedDistribution);
            // remove the cluster from the sorted list if its position changed
            if ((cluster->prev!=0 && cluster->ninstances < cluster->prev->ninstances)
             || (cluster->next!=0 && cluster->ninstances > cluster->next->ninstances)) {
                // either maintain sorted linkage, or maintain first/last elements
                if (cluster->prev) cluster->prev->next = cluster->next;
                else firstCluster = cluster->next;
                if (cluster->next) cluster->next->prev = cluster->prev;
                else lastCluster = cluster->prev;
                wasMoved = true;
            }
        }

        // insert the cluster at its new place if necessary
        if (wasMoved) {
            // re-insert it at its place
            if (!firstCluster) {
                cluster->prev = cluster->next = 0;
                lastCluster = firstCluster = cluster;
            }
            else if (cluster->ninstances < firstCluster->ninstances) {
                cluster->next = firstCluster;
                cluster->prev = 0;
                firstCluster->prev = cluster;
                firstCluster = cluster;
            }
            else {
                Cluster* c = firstCluster;
                for (; c->next!=0; c = c->next) if (cluster->ninstances < c->next->ninstances) break;
                cluster->next = c->next;
                cluster->prev = c;
                if (c->next) c->next->prev = cluster;
                else lastCluster = cluster;
                c->next = cluster;
            }
        }

        // update total & log
        totalInstances += deltaObs;
        logTotalInstances = log2(totalInstances);

        return cluster;
    }


    ~CausalStateAnalyzer() {
        while (firstCluster!=0) {
            Cluster* next = firstCluster->next;
            delete firstCluster;
            firstCluster = next;
        }
    }


// USER API BELOW

    CausalStateAnalyzer(double _significanceLevel, bool isChiElseKS = true) : significanceLevel(_significanceLevel), tester(isChiElseKS), totalInstances(0) {
        lastCluster = firstCluster = 0;
    }

    // Add a new observed future light cone for a given past light cone
    // This will update the plc distibution and the clusters and their complexity
    // The (potentially new) cluster for that plc is returned.
    Cluster* addObservation(PastLightCone plc, FutureLightCone flc) {
        return add_or_remove_Observation<&FutureDistribution::insert,1>(plc,flc);
    }

    // Useful for systems changing over time or space. When "old" observations
    // have a different enough distribution from current observation (non-stationary system)
    // and should be discarded.
    // If there are still instances for that plc, the corresponding cluster is returned, otherwise 0.
    Cluster* removeObservation(PastLightCone plc, FutureLightCone flc) {
        return add_or_remove_Observation<&FutureDistribution::remove,-1>(plc,flc);
    }

    // get absolute cluster complexity
    double getComplexity(Cluster* cluster) {
        return logTotalInstances - cluster->logNinstances;
    }

    // scale complexity to 0..1 relatively to the other clusters
    double getScaledComplexity(Cluster* cluster) {
        double minLog = firstCluster->logNinstances;
        double maxLog = lastCluster->logNinstances;
        double scaleFactor = maxLog - minLog;
        if (scaleFactor!=0.0) scaleFactor = 1.0 / scaleFactor;
        return (maxLog - cluster->logNinstances) * scaleFactor;
    }

    // TODO: make this incremental
    double getGlobalComplexity() {
        double globalComplexity = 0.0;
        for (Cluster* c = firstCluster; c!=0; c = c->next) {
            globalComplexity += c->ninstances * c->logNinstances;
        }
        return logTotalInstances - globalComplexity / totalInstances;
    }
};

